
#include <cstring>

#include <mmintrin.h>
#include <xmmintrin.h>

static inline void mm_empty() {_mm_empty();}

struct v8byte {
  __m64 d;
  v8byte() {}
  v8byte(__m64 d0) : d(d0) {}
  inline v8byte & zero() {
    d = _mm_setzero_si64(); return *this;}
  inline v8byte & ones() {
    d = _mm_set1_pi8(-1); return *this;}
  inline v8byte & set(unsigned char b) {
    d = _mm_set1_pi8(b); return *this;}
  inline v8byte & operator+= (v8byte y) {
    d = _m_paddb(d,y.d); return *this;}
  inline v8byte & operator-= (v8byte y) {
    d = _m_psubb(d,y.d); return *this;}
};
struct v8ubyte : public v8byte 
{
  v8ubyte() {}
  v8ubyte(__m64 d0) : v8byte(d0) {}
  v8ubyte(v8byte v0) : v8byte(v0) {}
};
struct v8sbyte : public v8byte 
{
  v8sbyte() {}
  v8sbyte(__m64 d0) : v8byte(d0) {}
  v8sbyte(v8byte v0) : v8byte(v0) {}
};

static inline v8ubyte min(v8ubyte x, v8ubyte y) {
  return _m_pminub(x.d,y.d);}
static inline v8ubyte max(v8ubyte x, v8ubyte y) {
  return _m_pmaxub(x.d,y.d);}
static inline v8ubyte avg(v8ubyte x, v8ubyte y) {
  return _m_pavgb(x.d,y.d);}

static inline v8byte operator+ (v8byte x, v8byte y) {
  return _m_paddb(x.d,y.d);}
static inline v8byte operator- (v8byte x, v8byte y) {
  return _m_psubb(x.d,y.d);}

static inline v8byte sadd (v8ubyte x, v8ubyte y) {
  return _m_paddusb(x.d,y.d);}
static inline v8byte ssub (v8ubyte x, v8ubyte y) {
  return _m_psubusb(x.d,y.d);}

static inline v8byte operator== (v8byte x, v8byte y) {
  return _m_pcmpeqb(x.d,y.d);}

static inline v8sbyte operator> (v8sbyte x, v8sbyte y) {
  return _m_pcmpgtb(x.d,y.d);}
static inline v8sbyte operator< (v8sbyte x, v8sbyte y) {
  return y > x;}

static inline v8byte operator& (v8byte x, v8byte y) {
  return _m_pand(x.d,y.d);}
static inline v8byte andn (v8byte x, v8byte y) {
  return _m_pandn(x.d,y.d);}
static inline v8byte operator| (v8byte x, v8byte y) {
  return _m_por(x.d,y.d);}
static inline v8byte operator^ (v8byte x, v8byte y) {
  return _m_pxor(x.d,y.d);}

static inline v8ubyte operator>= (v8ubyte x, v8ubyte y) {
  return min(x,y) == y;}
static inline v8ubyte operator<= (v8ubyte x, v8ubyte y) {
  return max(x,y) == y;}

static inline v8ubyte dist(v8ubyte x, v8ubyte y) {
  return ssub(x,y) | ssub(y,x);}

static inline v8byte if_else(v8byte p, v8byte x, v8byte y) {
  return (x & p) | andn(p,y);}

static inline v8sbyte operator-(v8sbyte x) {
  return v8byte().zero() - x;}
